#include <TaskThread.hpp>
void ConcurrentQueue::push(struct ibv_wc *const &data)
{
    std::unique_lock<std::mutex> lock(queueMutex);
    std::cout << "putting data\n";
    queue1.push(data);
    lock.unlock();
    queueCv.notify_one();
}

bool ConcurrentQueue::empty()
{
    std::unique_lock<std::mutex> lock(queueMutex);
    return queue1.empty() && queue2.empty();
}

struct ibv_wc *ConcurrentQueue::try_pop()
{
    struct ibv_wc *value = NULL;
    std::unique_lock<std::mutex> lock(queueMutex);
    /* first check in queue 2 if it is empty then check in queue 1
     * Since queue2 is empty and if queue1 is empty we can make thread to wait
     * Currently thread will busy wait if there are contending request in queue2
     * ie same request is in thread 2 and same request exist in runningRequests set
     */
    if (queue2.empty())
    {
        queueCv.wait(lock, [&]
                     { return queue1.size() > 0; });
        value = queue1.front();
        queue1.pop();
    }
    else
    {
        value = queue2.front();
        queue2.pop();
    }
    /* At this point value will not be null because it either gets data
     * from queue 1 or form queue2
     * To make it sequential consistent we need to order only put requests
     * it means ,we can return all other requests immediately
     */
    MessageHeader *request = (MessageHeader *)value->wr_id;
    if (value->opcode == IBV_WC_RECV && request->type == MessageType::PUT)
    {
        //  std::cout<<"value "<<value<<std::endl;
        /* Since this is the first request we need to put this in runningRequests set
         * because this might contend in future and return this immediately
         */
        if (runningRequests.empty())
        {
            runningRequests.insert(value);
            return value;
        }
        /* If this is contending with some runningRequest we need to put it inside queue2
         * and check for other request till queue1 not become empty, if there is no request in
         * queue1 then we will return null and taskthread will retry(probable busy waiting scenario)
         * Currently this way will make busy waiting of thread if queue1 is empty,
         * and it will always check in queue2 whenever it will retry for next time
         * probable solution is to add cv and wait it until we consume that runningRequest or new request come in queue1
         */
        auto it = runningRequests.find(value);
        if (it != runningRequests.end())
        {
            queue2.push(value);
            // std::cout<<"found putting in 2"<<std::endl;
            return NULL;
        }
        return value;
    }
    /* Returning all other request immediately*/
    return value;
}
void ConcurrentQueue::removeFromSet(struct ibv_wc *data)
{
    /* Remove Request from runningRequest Set after processing it so that other threads can 
     * process contending request from queue2
     */
    std::unique_lock<std::mutex> lock(queueMutex);
    // std::cout<<"removing"<<data<<std::endl;
    runningRequests.erase(data);
}
/* needs correction*/
void ConcurrentQueue::wait_and_pop(struct ibv_wc *&popped_value)
{
    std::unique_lock<std::mutex> lock(queueMutex);
    queueCv.wait(lock, [&]
                 { return queue1.size() > 0; });
    popped_value = queue1.front();
    queue1.pop();
}