#include "RdmaClientRepEndpoint.hpp"


int RdmaClientRepEndpoint::CONN_STATE_INITIALIZED = 0;
int RdmaClientRepEndpoint::CONN_STATE_ADDR_RESOLVED = 1;
int RdmaClientRepEndpoint::CONN_STATE_ROUTE_RESOLVED = 2;
int RdmaClientRepEndpoint::CONN_STATE_RESOURCES_ALLOCATED = 3;
int RdmaClientRepEndpoint::CONN_STATE_CONNECTED = 4;
int RdmaClientRepEndpoint::CONN_STATE_PARTIAL_CLOSED = 5;
int RdmaClientRepEndpoint::CONN_STATE_CLOSED = 6;

RdmaClientRepEndpoint::RdmaClientRepEndpoint(struct rdma_cm_id *id, int sendQueueSize, int recvQueueSize,
                                 int sendMsgSize, int recvMsgSize, int maxInLine, int timeout, rocksdb::DB *db)
    : RdmaRepEndpoint(id, sendQueueSize, recvQueueSize, sendMsgSize,
      recvMsgSize,db), _maxInLine(maxInLine), _timeoutMs(timeout)
{
    _state = CONN_STATE_INITIALIZED;
}

void RdmaClientRepEndpoint::connect(const char *ip, const char *port, const char *connData)
{
    _connData = connData;
    if (_state != CONN_STATE_INITIALIZED)
    {
        CPPLog::LOG_ERROR("RdmaClientRepEndpoint : connect state not initialized");
    }
    int ret;
    CPPLog::LOG_DEBUG("RdmaClientRepEndpoint :  step2 getaddrinfo");
    struct addrinfo *addr;
    ret = getaddrinfo(ip, port, NULL, &addr);
    if (ret)
    {
        CPPLog::LOG_ERROR("RdmaClientRepEndpoint : get_addr_info failed");
    }

    CPPLog::LOG_DEBUG("RdmaClientRepEndpoint :  step2 resolve addr");
    ret = rdma_resolve_addr(_cm_id, NULL, addr->ai_addr, _timeoutMs);
    if (ret)
    {
        CPPLog::LOG_ERROR("unable to resolve addr");
        //return;
    }
    CPPLog::LOG_DEBUG("RdmaClientRepEndpoint :  step2 resolve addr resolved");
    _state = CONN_STATE_ADDR_RESOLVED;
}

bool RdmaClientRepEndpoint::isConnected()
{
    return _state == CONN_STATE_CONNECTED;
}

void RdmaClientRepEndpoint::processCmEvent(struct rdma_cm_event *event)
{
    if (event->event == RDMA_CM_EVENT_ADDR_RESOLVED && event->id != NULL)
    {
        CPPLog::LOG_DEBUG("RdmaClientRepEndpoint : step3 resolve_route");
        rdma_resolve_route(_cm_id, _timeoutMs);
    }
    else if (event->event == RDMA_CM_EVENT_ROUTE_RESOLVED && event->id != NULL)
    {
        registerMemory();
        CPPLog::LOG_DEBUG("RdmaClientRepEndpoint : step5 connect");
        connect();
    }
    else if (event->id != NULL && event->event == RDMA_CM_EVENT_ESTABLISHED)
    {
        CPPLog::LOG_DEBUG("RdmaClientRepEndpoint : step6 Connected");
        _state = CONN_STATE_CONNECTED;
    }
    else if (event->id != NULL && event->event == RDMA_CM_EVENT_UNREACHABLE)
    {
        std::cout << "Not able to connect replication endpoint\n";
        if(_retryCount < _maxRetryCount)
        {
            std::cout<<"retry to connect with server\n";
            connect();
            _retryCount ++;
        }
        
    }
    else if (event->id != NULL && event->event == RDMA_CM_EVENT_DISCONNECTED)
    {
        CPPLog::LOG_DEBUG("RdmaClientRepEndpoint : step7 Closed");
        close();
    }
    else
    {
        std::ostringstream ss;
        ss << "RdmaClientRepEndpoint : Not able to procces CM EVent" << rdma_event_str(event->event);
        ss << event->id << " " << event->listen_id;
        CPPLog::LOG_ERROR(ss);
    }
}

void RdmaClientRepEndpoint::close()
{
    if (_state != CONN_STATE_CONNECTED)
    {
        std::cout << "RdmaClientRepEndpoint : close invalid state" << std::endl;
    }
    _state = CONN_STATE_PARTIAL_CLOSED;
    int ret = rdma_disconnect(_cm_id);
    if (ret)
    {
        std::cout << "RdmaClientRepEndpoint : rdma_disconnect failed" << std::endl;
    }
    delete[](char *) _sendBuff;
    delete[](char *) _recvBuff;
    rdma_dereg_mr(_sendMr);
    rdma_dereg_mr(_recvMr);
    rdma_destroy_qp(_cm_id);
    // rdma_destroy_id(_cm_id);
}

void RdmaClientRepEndpoint::connect()
{
    if (_connData != NULL)
    {
        struct rdma_conn_param conn_param;
        memset(&conn_param, 0, sizeof(conn_param));
        conn_param.responder_resources = 1;
        conn_param.initiator_depth = 1;
        conn_param.retry_count = 7;
        conn_param.rnr_retry_count = 7;
        conn_param.private_data = _connData;
        conn_param.private_data_len = strlen(_connData);
        rdma_connect(_cm_id, &conn_param);
    }
    else
    {
        rdma_connect(_cm_id, NULL);
    }
}

void RdmaClientRepEndpoint::registerMemory()
{
    if (_state != CONN_STATE_ROUTE_RESOLVED)
    {
        std::cout << "RdmaClientRepEndpoint : createResource address not resolved" << std::endl;
        return;
    }
    _sendBuff = new char[_sendMsgSize * _sendQueueSize];
    if (_sendBuff == NULL)
    {
        std::cout << "RdmaClientRepEndpoint :  sendBuff allocation failed" << std::endl;
        return;
    }

    _sendMr = rdma_reg_msgs(_cm_id, _sendBuff, _sendMsgSize * _sendQueueSize);
    if (_sendMr == NULL)
    {
        std::cout << "RdmaClientRepEndpoint :  sendMr reg failed" << std::endl;
        return;
    }

    _recvBuff = new char[_recvMsgSize * _recvQueueSize];
    if (_recvBuff == NULL)
    {
        std::cout << "RdmaClientRepEndpoint :  recvBuff allocation failed" << std::endl;
        return;
    }

    _recvMr = rdma_reg_msgs(_cm_id, _recvBuff, _recvMsgSize * _recvQueueSize);
    if (_recvMr == NULL)
    {
        std::cout << "RdmaClientRepEndpoint :  recvMr reg failed" << std::endl;
        return;
    }
    char *buffer = (char *)_recvBuff;
    for (int i = 0; i < _recvQueueSize; i++)
    {
        void *const location = buffer + i * _recvMsgSize;
        rdma_post_recv(_cm_id, reinterpret_cast<void *>(location), reinterpret_cast<void *>(location),
                       _recvMsgSize, _recvMr);
    }
    buffer = (char *)_sendBuff;
    for (int i = 0; i < _sendQueueSize; i++)
    {
        void *const location = buffer + i * _sendMsgSize;
        std::unique_lock<std::mutex> lock(_sendBuffersM);
        _sendBuffers.push((char *)location);
    }

    _state = CONN_STATE_RESOURCES_ALLOCATED;
}

void RdmaClientRepEndpoint::createResources(struct ibv_cq *cq)
{
    if (_state != CONN_STATE_ADDR_RESOLVED)
    {
        std::cout << "RdmaClientRepEndpoint : createResource address not resolved" << std::endl;
        return;
    }
    _protectionDomain = ibv_alloc_pd(_cm_id->verbs);
    if (_protectionDomain == NULL)
    {
        std::cout << "RdmaClientRepEndpoint : ibv_alloc_pd failed " << std::endl;
        return;
    }
    struct ibv_cq *completionQueue = cq;
    struct ibv_qp_init_attr qp_init_attr;
    memset(&qp_init_attr, 0, sizeof(qp_init_attr));
    // This is used to set endpoint address with qp
    qp_init_attr.qp_context = (void *)this;
    // if not set 0, all work requests submitted to SQ will always generate a Work Completion
    qp_init_attr.sq_sig_all = 1;
    // completion queue can be shared or you can use distinct completion queues.
    qp_init_attr.send_cq = completionQueue;
    qp_init_attr.recv_cq = completionQueue;
    qp_init_attr.qp_type = IBV_QPT_RC;
    // increase if you want to keep more send work requests in the SQ.
    qp_init_attr.cap.max_send_wr = _sendQueueSize;
    // increase if you want to keep more receive work requests in the RQ.
    qp_init_attr.cap.max_recv_wr = _recvQueueSize;
    // increase if you allow send work requests to have multiple scatter gather entry (SGE).
    qp_init_attr.cap.max_send_sge = 1;
    // increase if you allow receive work requests to have multiple scatter gather entry (SGE).
    qp_init_attr.cap.max_recv_sge = 1;
    qp_init_attr.cap.max_inline_data = _maxInLine;

    //_queuePair = ibv_create_qp(_protectionDomain, &qp_init_attr);
    int ret = rdma_create_qp(_cm_id, _protectionDomain, &qp_init_attr);
    if (ret)
    {
        std::cout << "RdmaClientRepEndpoint : ibv_create_cq failed\n";
    }
    if (_cm_id->pd == NULL)
    {
        std::cout << "RdmaClientRepEndpoint : pd not set" << std::endl;
        _cm_id->pd = _protectionDomain;
    }

    _state = CONN_STATE_ROUTE_RESOLVED;
}

