#include "RdmaServerEndpointGroup.hpp"

int RdmaServerEndpointGroup::CONN_STATE_INITIALIZED = 0;
int RdmaServerEndpointGroup::CONN_STATE_BINDED = 3;
int RdmaServerEndpointGroup::CONN_STATE_CONNECTED = 4;
int RdmaServerEndpointGroup::CONN_STATE_CLOSED = 5;

RdmaServerEndpointGroup::RdmaServerEndpointGroup(int sendQueueSize, int recvQueueSize, int compQueueSize,
                                                 int sendMsgSize, int recvMsgSize, rocksdb::DB *db)
    : _sendQueueSize(sendQueueSize), _recvQueueSize(recvQueueSize), _compQueueSize(compQueueSize),
      _sendMsgSize(sendMsgSize), _recvMsgSize(recvMsgSize), _db(db)
{
    CPPLog::LOG_INFO("SalEndpointGroup : Step 1 creating event channel");
    _eventChannel = rdma_create_event_channel();
    if (_eventChannel == NULL)
    {
        CPPLog::LOG_ERROR("SalEndpointGroup : error creating event channel");
    }
    int ret = rdma_create_id(_eventChannel, &_cm_id, NULL, RDMA_PS_TCP);
    if (ret == -1)
    {
        CPPLog::LOG_ERROR("CMProcesor : rdma_create_id failed");
        exit(-1);
    }
    _qpSalEndpointMap = new std::unordered_map<uint32_t, RdmaSalEndpoint *>();
    _qpRepEndpointMap = new std::unordered_map<uint32_t, RdmaRepEndpoint *>();
}
void RdmaServerEndpointGroup::setExecutor(Executor *ex)
{
    _executor = ex;
}
void RdmaServerEndpointGroup::startCmProcessor(bool newThread)
{
    if (newThread == true)
        _cmEventThread = new std::thread(&RdmaServerEndpointGroup::processCmEvents, this);
    else
        processCmEvents();
}
void RdmaServerEndpointGroup::processCmEvents()
{
    int ret;
    struct rdma_cm_event *event;
    CPPLog::LOG_INFO("SalEndpointGroup : starting cm processing thread");
    while (!_stopCmProcessor)
    {
        ret = rdma_get_cm_event(_eventChannel, &event);
        if (ret)
        {
            CPPLog::LOG_ERROR("SalEndpointGroup : rdma_get_cm_event failed");
            continue;
        }
        processCmEvent(event);
        ret = rdma_ack_cm_event(event);
        if (ret)
        {
            CPPLog::LOG_ERROR("SalEndpointGroup : rdma_ack_cm_event failed");
        }
    }
}

void RdmaServerEndpointGroup::bind(const char *ip, const char *port, int backlog)
{
    int ret;
    std::ostringstream ss;
    ss<<"RdmaServerEndpointGroup : Step 2 bind_addr"<<ip<<" "<<port<<"\n";
    CPPLog::LOG_ALWAYS(ss);
    struct addrinfo *addr;
    ret = getaddrinfo(ip, port, NULL, &addr);
    if (ret)
    {
        CPPLog::LOG_ERROR("RdmaServerEndpointGroup : get_addr_info failed");
    }
    ret = rdma_bind_addr(_cm_id, addr->ai_addr);
    if (ret)
    {
        CPPLog::LOG_ERROR("RdmaServerEndpointGroup : rdma_bind_addr failed");
        return;
    }

    CPPLog::LOG_ALWAYS("RdmaServerEndpointGroup : rdma_bind_addr successful");
    ret = rdma_listen(_cm_id, backlog);
    if (ret)
    {
        CPPLog::LOG_ERROR("RdmaServerEndpointGroup : rdma_listen failed");
        return;
    }
}

struct ibv_cq *RdmaServerEndpointGroup::createCq(struct rdma_cm_id *id)
{
    if (_cqProcessor == NULL)
    {
        CPPLog::LOG_ALWAYS("RdmaServerEndpointGroup : step 5 create cq processor");
        _cqProcessor = new RdmaCqProcessor(_executor, _cm_id->verbs, _compQueueSize);
        _cqProcessor->start();
    }
    return _cqProcessor->getCq();
}


void RdmaServerEndpointGroup::createEpCmEvent(struct rdma_cm_event *event)
{
    CPPLog::LOG_ALWAYS("RdmaServerEndpointGroup : step 4 Got Connect Request Sal Endpoint");
    /*
     * create and add to vectors for replication and invalidation processing
     * connData is used to identify whether connection came for client or from other Sal;
     */

    const char *connData = reinterpret_cast<const char *>(event->param.conn.private_data);

    if (strcmp(connData, "sal") == 0)
    {
        RdmaSalEndpoint *endpoint = nullptr;
        endpoint = new RdmaSalEndpoint(event->id, createCq(event->id), _sendQueueSize, _recvQueueSize,
                                       _sendMsgSize, _recvMsgSize, _db);
        event->id->context = (void *)endpoint;
        endpoint->createResources();
        // std::unique_lock lock(_salMutex);
        std::cout<<"Sal Map Size"<<_qpSalEndpointMap->size()<<"qp "<<event->id->qp->qp_num<<"\n";
        _qpSalEndpointMap->emplace(event->id->qp->qp_num, endpoint);

        std::cout<<"Sal Map Size"<<_qpSalEndpointMap->size()<<"\n";
    }
    else
    {
        RdmaServerRepEndpoint *endpoint = nullptr;
        endpoint = new RdmaServerRepEndpoint(event->id, createCq(event->id), _sendQueueSize, _recvQueueSize,
                                             _sendMsgSize, _recvMsgSize, _db);
        event->id->context = (void *)endpoint;
        endpoint->createResources();
        // std::unique_lock lock(_salMutex);
        _qpRepEndpointMap->emplace(event->id->qp->qp_num, endpoint);
    }

    rdma_accept(event->id, NULL);
}

void RdmaServerEndpointGroup::processCmEvent(struct rdma_cm_event *event)
{
    std::ostringstream ss;
    ss << "RdmaServerEndpointGroup : event " << rdma_event_str(event->event);
    ss << " id " << event->id;
    CPPLog::LOG_ALWAYS(ss);
    /*
     * Connect request came on listener ie endpointgroup
     * in this listen_id contains cm_id of listener and id contains the new id of connection request
     */
    if (event->event == RDMA_CM_EVENT_CONNECT_REQUEST && event->listen_id != NULL)
    {
        createEpCmEvent(event);
    }
    else if (event->id != NULL && _cm_id == event->id)
    {
        if (event->event == RDMA_CM_EVENT_DISCONNECTED)
        {
            std::cout << "RdmaServerEndpointGroup : Disconnect Sal Endpoint" << std::endl;
            close();
        }
        else
        {
            std::cout << "RdmaServerEndpointGroup : unknown Event for listener" << rdma_event_str(event->event) << std::endl;
        }
    }
    /*
     * Event came for a endpoint
     */
    else if (event->id != NULL && event->id->context != NULL)
    {
        /*
         * TIMEWAIT_EXIT happens only when client is already closed and we are not able to send
         * rdma_disconnect()
         * Since endpoint already closed with RDMA_CM_EVENT_DISCONNECTED
         * we need to return for this event
         */
        if (event->event == RDMA_CM_EVENT_TIMEWAIT_EXIT)
        {
            CPPLog::LOG_ALWAYS("Recieved TIMEWAIT_EXIT");
            return;
        }

        uint32_t qp = event->id->qp->qp_num;
        auto it = _qpSalEndpointMap->find(qp);
        if (it != _qpSalEndpointMap->end())
        {
            ((RdmaSalEndpoint *)event->id->context)->processCmEvent(event);
        }
        auto it2 = _qpRepEndpointMap->find(qp);
        if (it2 != _qpRepEndpointMap->end())
        {
            ((RdmaServerRepEndpoint *)event->id->context)->processCmEvent(event);
        }

        if (event->event == RDMA_CM_EVENT_DISCONNECTED)
        {
            // std::unique_lock lock(_salMutex);
            auto it = _qpSalEndpointMap->find(qp);
            if (it != _qpSalEndpointMap->end())
            {
                _qpSalEndpointMap->erase(qp);
                delete ((RdmaSalEndpoint *)event->id->context);
            }
            auto it2 = _qpRepEndpointMap->find(qp);
            if (it2 != _qpRepEndpointMap->end())
            {
                _qpRepEndpointMap->erase(qp);
                delete ((RdmaServerRepEndpoint *)event->id->context);
            }
        }
    }
    else
    {
        std::ostringstream ss;
        ss << "RdmaServerEndpointGroup :  Not able to procces CM EVent";
        ss << rdma_event_str(event->event) << event->id << " ";
        ss << event->listen_id << std::endl;
        CPPLog::LOG_ERROR(ss);
    }
}

void RdmaServerEndpointGroup::close()
{
    _cqProcessor->close();
    //_executor->stop();
    delete _cqProcessor;
    CPPLog::LOG_ALWAYS("Closing CM Processor");
    _stopCmProcessor = true;
    if (_cmEventThread != NULL)
        _cmEventThread->join();
    delete _cmEventThread;
    rdma_destroy_event_channel(_eventChannel);

    for (auto it = _qpSalEndpointMap->begin(); it != _qpSalEndpointMap->end(); it++)
    {
        ((RdmaSalEndpoint *)it->second)->close();
        delete ((RdmaSalEndpoint *)it->second);
    }
    delete _qpSalEndpointMap;
    
    for (auto it = _qpRepEndpointMap->begin(); it != _qpRepEndpointMap->end(); it++)
    {
        ((RdmaServerRepEndpoint *)it->second)->close();
        delete ((RdmaServerRepEndpoint *)it->second);
    }
    delete _qpRepEndpointMap;
    rdma_disconnect(_cm_id);
    rdma_destroy_id(_cm_id);
}