#include "TaskThread.hpp"

TaskThread::TaskThread(int id, int cpu, ConcurrentQueue *taskqueue,
                       std::unordered_map<uint32_t, RdmaRepEndpoint *> *clientRepMap,
                       std::unordered_map<uint32_t, RdmaRepEndpoint *> *serverRepMap,
                       std::unordered_map<uint32_t, RdmaSalEndpoint *> *salMap)
    : _id(id), _clientRepMap(clientRepMap), _serverRepMap(serverRepMap), _salMap(salMap)
{
    _taskQueue = taskqueue;
    if (pthread_create(&thread, NULL, &TaskThread::run, this))
    {
        CPPLog::LOG_ERROR("pthread create has been failed while creating taskthread");
        exit(0);
    }
    /*cpu_set_t cpuset;
    CPU_ZERO(&cpuset);
    CPU_SET(cpu, &cpuset);
    std::ostringstream ss;
    ss << "New Thread Setting CPU affinty " << cpu;
    CPPLog::LOG_ALWAYS(ss);
    if (pthread_setaffinity_np(thread, sizeof(cpu_set_t), &cpuset) != 0)
    {
        CPPLog::LOG_ERROR("Error calling pthread_setaffinity_np ");
    }*/
    pthread_setname_np(thread, "TaskThread");
}

TaskThread::~TaskThread()
{
    CPPLog::LOG_INFO("TaskThread Destructed");
    stop();
}

void TaskThread::stop()
{
    _stop = true;
    if (pthread_join(thread, NULL) == 0)
    {
        CPPLog::LOG_ERROR("pthread join failed");
    }
}

inline void *TaskThread::run(void *object)
{
    /*
     * This is because data from pthread_create is passed as void*
     */
    TaskThread *thread = reinterpret_cast<TaskThread *>(object);
    std::ostringstream ss;
    ss<<"Running task thread"<<thread->_id;
    CPPLog::LOG_ALWAYS(ss);
    while (!thread->_stop)
    {
        struct ibv_wc *data = NULL;
        std::cout << "Get start\n";
        data = thread->_taskQueue->try_pop();
        if (data != NULL)
        {
            std::cout << "TaskThread:: got data";
            if (data == NULL || data->status != IBV_WC_SUCCESS)
            {
                std::ostringstream ss;
                ss << "TaskThread : failed work completion : ";
                ss << ibv_wc_status_str(data->status) << " on qp " << data->qp_num;
                CPPLog::LOG_ERROR(ss);
                continue;
            }
            RdmaSalEndpoint *salEp = nullptr;
            RdmaRepEndpoint *repEp = nullptr;
            auto it = thread->_salMap->find(data->qp_num);
            if (it != thread->_salMap->end())
            {
                salEp = it->second;
            }
            auto it2 = thread->_clientRepMap->find(data->qp_num);
            if (it2 != thread->_clientRepMap->end())
            {
                repEp = it2->second;
            }
            else
            {
                auto it3 = thread->_serverRepMap->find(data->qp_num);
                if(it3 != thread->_serverRepMap->end())
                    repEp = it3->second;
            }
            if (salEp == nullptr && repEp == nullptr)
            {
                std::ostringstream ss;
                ss<<"RdmaSal : endpoint not registered for qp"<<data->qp_num;
                CPPLog::LOG_INFO(ss);
                
                continue;
            }
            else if (salEp != nullptr)
                thread->processEvent(salEp, data);
            else if (repEp != nullptr)
                thread->processRepEvent(repEp, data);
            thread->_taskQueue->removeFromSet(data);
            delete data;
        }
    }
    return NULL;
}

void TaskThread::processEvent(RdmaSalEndpoint *ep, struct ibv_wc *data)
{
    std::cout<<"processing sal event\n";
    /* sal Request*/
    switch (data->opcode)
    {
    case IBV_WC_SEND:
        ep->processSendCompletion(data);
        break;
    case IBV_WC_RECV:
    {
        // it->second->processRecvCompletion(data);
        char *buffer = new char[data->byte_len];
        memcpy(buffer, (void *)data->wr_id, data->byte_len);
        rdma_post_recv(ep->_cm_id, (void *)data->wr_id, (void *)data->wr_id,
                       ep->_recvMsgSize, ep->_recvMr);
        struct MessageHeader *req = (struct MessageHeader *)buffer;
        std::cout << "TaskThread 1\n";
        switch (req->type)
        {
        case MessageType::GET:
            ep->processGet(req);
            break;
        case MessageType::DELETE:
                replicateSalRequest(buffer, data->byte_len);
            ep->processDelete(req);
            break;
        case MessageType::PUT:
                replicateSalRequest(buffer, data->byte_len);
            ep->processPut(req);
            break;
        default:
            std::ostringstream ss;
            ss<<"SalRequest invalid req type"<<data->opcode;
            CPPLog::LOG_ERROR(ss);
            break;
        }
        delete[] buffer;
    }
    break;
    default:
        std::ostringstream ss;
        ss << "TaskThread default opcode : " << data->opcode;
        CPPLog::LOG_INFO(ss);
        break;
    }
}

void TaskThread::processRepEvent(RdmaRepEndpoint *ep, struct ibv_wc *data)
{
    switch (data->opcode)
    {
    case IBV_WC_SEND:
        ep->processSendCompletion(data);
        break;
    case IBV_WC_RECV:
    {
        // it->second->processRecvCompletion(data);
        char *buffer = new char[data->byte_len];
        memcpy(buffer, (void *)data->wr_id, data->byte_len);
        rdma_post_recv(ep->_cm_id, (void *)data->wr_id, (void *)data->wr_id,
                       ep->_recvMsgSize, ep->_recvMr);
        struct MessageHeader *req = (struct MessageHeader *)buffer;
        std::cout << "TaskThread 1\n";
        switch (req->type)
        {
        case MessageType::GET:
            ep->processGet(req);
            break;
        case MessageType::DELETE:
            ep->processDelete(req);
            break;
        case MessageType::PUT:
            ep->processPut(req);
            break;
        default:
            CPPLog::LOG_ERROR("SalRequest invalid req type");
            break;
        }
        delete[] buffer;
    }
    break;
    default:
        std::ostringstream ss;
        ss << "TaskThread default opcode : " << data->opcode;
        CPPLog::LOG_INFO(ss);
        break;
    }
}

void TaskThread::replicateSalRequest(char *req, uint32_t size)
{
    auto repIt = _clientRepMap->begin();
    for (; repIt != _clientRepMap->end(); repIt++)
    {
        repIt->second->sendMessage(req, size);
    }
    auto serverRepIt = _serverRepMap->begin();
    for (; serverRepIt != _serverRepMap->end(); serverRepIt++)
    {
        serverRepIt->second->sendMessage(req, size);
    }

    MessageHeader *salReq = (MessageHeader *)req;
    char *buffer = new char[MessageHeaderSize + salReq->keySize];
    MessageHeader *invRequest = (MessageHeader *)(buffer);
    invRequest->type = MessageType::INVALIDATE;
    invRequest->id = salReq->id;
    invRequest->keySize = salReq->keySize;
    memcpy(buffer + MessageHeaderSize, salReq + MessageHeaderSize, salReq->keySize);
    // Send Invalidation to sal's
    auto salIt = _salMap->begin();
    for (;salIt != _salMap->end();salIt++)
    {
        salIt->second->sendMessage(buffer, MessageHeaderSize + salReq->keySize);
    }
    delete[] buffer;
}